from optimpy.utils.helpers import *
from tqdm import tqdm
import numpy as np
import copy 



class TabuList:
    def __init__(self, timer: int):
        self.__TB = [[None,None,-1, 100000000]]
        self.timer = timer
        
    def __str__(self):
        printable_object = "\n Tabu list: \n"
        printable_object += "List size: {} \n".format(len(self.__TB))
        for i in range(len(self.__TB)):
            a,b,c,d = self.__TB[i][0], self.__TB[i][1], self.__TB[i][2], self.__TB[i][3]
            printable_object += \
            "p: {} v:{} Iteration: {} current timer: {} \n".format(a,b,c,d)
        return printable_object
        
    @checkargs
    def push_front(self, x: list) -> None:
        assert len(x) == 3
        x.append(self.timer)
        self.__TB.insert(0,x)
    
    def pop_back(self) -> None:
        x_last = self.__TB.pop()
        x_current_back = self.get_back()
        update_time = self.timer - (x_current_back[2]-x_last[2])
        x_current_back[2] -= update_time
        self.update_back(x_current_back)
        
    @checkargs
    def update_back(self, x : list) -> None:
        self.__TB[-1] = x
    
    @checkargs
    def find(self, x : list) -> bool:
        #X is [p , v], where p is the position changed and v by new value.
        assert len(x) == 2
        for i in range(len(self.__TB)):
            if x[0] == self.__TB[i][0] and x[1] == self.__TB[i][1]:
                return True
        return False
        
    def get_back(self) -> list:
        return self.__TB[-1]
    
    @checkargs
    def reset(self, timer : int) -> None:
        self.__TB = [] 
        self.timer = timer


class TabuSearch:
    @checkargs
    def __init__(self,function: function_type,constraints : list):
        """
        Arguments: 
            - function: Objetive function to maximize.
            - Constraints: constraints to be a feasible solution.
        """
        self.TL = TabuList(2) #Initialize tabulist.
        self.Constraints = constraints
        self.f = function
        self.best_solution = None
        self.f_best = None
        
    def __str__(self):
        return "Tabú search: \n f(X) = {} \n X = {}".format(self.f_best,self.best_solution)
        
    @checkargs
    def optimize(self,X0: np.ndarray ,iterations: int , memory : int)->None:
        """
        ------------------------------------------------------
        Arguments:
            - X0: Numpy array, represent initial solution.
            - Iterations:  Integer, defining like stop condition.
            - Memory: Integer, time before erase that solution. 
        ------------------------------------------------------
        """
        self.TL.reset(memory)
        self.best_solution = None
        self.f_best = None
        
        best_candidate =copy.deepcopy(X0)
        f_candidate =self.f(best_candidate)
        
        for step_ in tqdm(range(1,iterations+1)):
        
            Neighbors =[neighbor for  neighbor in self.get_neighbors(best_candidate) \
                        if not self.TL.find(self.encode_change(neighbor, best_candidate))]
            Neighbors = np.array(Neighbors)
          
            ValidNeighbors =  \
            Neighbors[np.apply_along_axis(self.is_valid,1,Neighbors),:] #1
        
            if len(ValidNeighbors) == 0:
                continue
            
            f_feasible_candidates =  \
            np.apply_along_axis(self.f , 1, ValidNeighbors)      #2
 
            ind_min = np.argmin(f_feasible_candidates)           #3
             
            p,v = self.encode_change(ValidNeighbors[ind_min],best_candidate)
            self.TL.push_front([p, v, step_])
        
            if f_feasible_candidates[ind_min]< f_candidate:
                best_candidate = copy.deepcopy(ValidNeighbors[ind_min])
                f_candidate = f_feasible_candidates[ind_min]
            
            
            most_old_candidate = self.TL.get_back()
            most_old_candidate[-1] -= 1
            
            if(most_old_candidate[-1] <= 0): #4
                self.TL.pop_back()
            else:
                self.TL.update_back(most_old_candidate)
            
        self.best_solution = best_candidate
        self.f_best = f_candidate
    
    @checkargs
    def is_valid(self, x : np.ndarray) -> bool:
        for constrain in self.Constraints:
            if not constrain(x):
                return False
        return True
    
    def get_neighbors(self, x : (list,np.ndarray))-> list:
        raise NotImplementedError
        
        
    def encode_change(self, neighbor: (list,np.ndarray), x: (list,np.ndarray)) -> str:
        raise NotImplementedError
     